<?php

namespace Drupal\entity_sanitizer\Commands;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\entity_sanitizer\Sanitizer;
use Drush\Commands\DrushCommands;
use Drupal\entity_sanitizer\UnsupportedFieldTypeException;
use Drupal\Core\Entity\Sql\SqlEntityStorageInterface;

/**
 * A drush command file.
 *
 * @package Drupal\entity_sanitizer\Commands
 */
class EntitySanitizerCommands extends DrushCommands {

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected ConfigFactory $configFactory;

  /**
   * The sanitizer service.
   *
   * @var \Drupal\entity_sanitizer\Sanitizer
   */
  protected Sanitizer $sanitizer;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs a EntitySanitizerCommands object.
   *
   * @param \Drupal\entity_sanitizer\Sanitizer $sanitizer
   *   The sanitizer service.
   * @param \Drupal\Core\Config\ConfigFactory $config_factory
   *   The config factory service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(Sanitizer $sanitizer, ConfigFactory $config_factory, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct();
    $this->sanitizer = $sanitizer;
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Sanitize entity data.
   *
   * @command entity-sanitizer:sanitize
   * @bootstrap max
   * @aliases entity-sanitizer-sanitize, es
   *
   * @usage drush entity-sanitizer:sanitize
   *   Sanitizes all the non-whitelisted entities.
   */
  public function sanitizeNonWhitelistEntity() {
    // Get the whitelist configuration.
    $whitelist = $this->configFactory->get('entity_sanitizer.whitelist')->get('whitelist');

    // We retrieve the storage definitions for all fields stored in a database.
    // This automatically excludes any computed fields.
    $entities = $this->sanitizer->getAllEntityFieldDefinitions();

    // Our array with sanitizing operations.
    $sanitize_orders = [];

    echo "Sanitizing the following entity bundles: \n";

    // DEBUG: LIMIT TO A SINGLE ENTITY/BUNDLE FOR EASE!
    // $entities = ['node' => $entities['node']];
    // $entities['node'] = ['page' => $entities['node']['page']];.
    foreach ($entities as $entity_id => $bundles) {
      // Check for the entity to be skipped.
      if (isset($whitelist[$entity_id])
        && !is_array($whitelist[$entity_id])
        && array_key_exists($entity_id, $whitelist)
        ) {
        echo '------------------------------------------------' . PHP_EOL;
        echo "Skipping {$entity_id} due to whitelisting\n";
        echo '------------------------------------------------' . PHP_EOL;
        continue;
      }

      /** @var \Drupal\Core\Entity\Sql\SqlEntityStorageInterface $entity_storage */
      $entity_storage = $this->entityTypeManager->getStorage($entity_id);

      // We only know how to handle fields stored in SQL.
      if (!$entity_storage instanceof SqlEntityStorageInterface) {
        echo "Skipping {$entity_id} sanitizing because it's not stored in SQL.\n";
        continue;
      }

      echo "\n{$entity_id} \n";

      foreach ($bundles as $bundle_id => $fields) {

        // Check for the bundle to be skipped.
        // @todo Issue #3085448 Improve the logic to check existing keys for bundle.
        if ((array_key_exists($bundle_id, $whitelist[$entity_id] ?? [])) && !is_array($whitelist[$entity_id][$bundle_id])) {
          echo '------------------------------------------------' . PHP_EOL;
          echo "Skipping {$bundle_id} due to whitelisting\n";
          echo '------------------------------------------------' . PHP_EOL;
          continue;
        }

        echo "- {$bundle_id} \n";

        // Use the storage definitions to create a sanitizing order.
        /** @var \Drupal\Core\Field\FieldStorageDefinitionInterface $definition */
        foreach ($fields as $field_name => $definition) {
          // Check for the field to be skipped.
          if (isset($whitelist[$entity_id][$bundle_id]) && is_array($whitelist[$entity_id][$bundle_id]) && in_array($field_name, $whitelist[$entity_id][$bundle_id])) {
            echo '------------------------------------------------' . PHP_EOL;
            echo "Skipping {$field_name} due to whitelisting\n";
            echo '------------------------------------------------' . PHP_EOL;
            continue;
          }

          // Fields with custom storage have no table info so we skip them.
          if ($definition->hasCustomStorage()) {
            echo "Skipping field {$field_name} as it has custom storage.\n";
            continue;
          }

          // @todo Base fields don't have their own table so they need special
          //   attention.
          if ($definition->isBaseField()) {
            echo "Skipped base field {$field_name} ({$definition->getType()}).\n";
            continue;
          }

          // The type of field determines how we create the sanitized value.
          $field_type = $definition->getType();

          // Every field has a table for the non-revisioned version of the field.
          $revision_table_types = [FALSE];

          // Some also have a field revision table so they require more work.
          if ($entity_storage->getEntityType()->isRevisionable() && $definition->isRevisionable()) {
            $revision_table_types[] = TRUE;
          }

          foreach ($revision_table_types as $revision_table) {
            $table_name = $this->sanitizer->generateFieldTableName($definition, $revision_table);

            // Create an order to sanitize this table.
            if (!isset($sanitize_orders[$table_name])) {
              $sanitize_orders[$table_name] = [
                'field_name' => $field_name,
                'field_type' => $field_type,
                'is_revision_table' => $revision_table,
                'definition' => $definition,
                'bundles' => [],
              ];
            }

            // Add this bundle as one to be sanitized.
            $sanitize_orders[$table_name]['bundles'][] = $bundle_id;

            // @todo Remove this debug output.
            echo "-- {$table_name} ({$field_type})\n";
          }
        }
      }
    }

    echo PHP_EOL;

    // We generate the queries now so we can report any missing fields before
    // asking the user to continue.
    $queries = [];
    $unsupported_types = [];

    foreach ($sanitize_orders as $table_name => $order) {
      try {
        $queries[] = $this->sanitizer->generateSanitizeQuery($table_name, $order);
      }
      catch (UnsupportedFieldTypeException $exception) {
        $field_type = $exception->getFieldType();

        if (empty($unsupported_types[$field_type])) {
          $unsupported_types[$field_type] = [];
        }

        $unsupported_types[$field_type][] = $exception->getFieldName();
      }
    }

    if (!empty($unsupported_types)) {
      echo '------------------------------------------------' . PHP_EOL;
      echo 'The following field types were not supported and fields of those types will not be sanitized!' . PHP_EOL;
      foreach ($unsupported_types as $type => $fields) {
        $output_fields = implode(', ', $fields);
        echo PHP_EOL . "{$type} ({$output_fields})" . PHP_EOL;
      }
      echo '------------------------------------------------' . PHP_EOL . PHP_EOL;
    }

    // @todo need create UI for interaction
    //   if (!drush_confirm(dt('Do you really want to sanitize all these entities?'))) {
    //      return drush_user_abort();
    //    }
    foreach ($queries as $query) {
      // If we don't need a query for this order (e.g. entity reference) then we
      // receive NULL instead of a query to execute.
      if (!empty($query)) {
        $query->execute();
      }
    }

    // @todo Create a query for each field individually to update the value
    //   based on the field type.
    // @todo Add a WHERE condition to the UPDATE query to exclude whitelisted
    //   bundle/field combinations.
    //    foreach ($entities as $entity_id => $fields) {
    //      foreach ($fields as $field_name => $definition) {
    //      }
    //    }

    // @todo need add clearing cache drupal_flush_all_caches(); provides errors.
    echo PHP_EOL;

    $this->output()->writeln('Sanitization completed.');
  }

}
